#include "config_loader.h"

#include <LittleFS.h>
#include <ArduinoJson.h>

Config loadConfig() {
  Config config;

  if (!LittleFS.begin()) {
    return config;
  }

  File configFile = LittleFS.open("/config.json", "r");
  if (!configFile) {
    return config;
  }

  StaticJsonDocument<1024> doc;
  DeserializationError error = deserializeJson(doc, configFile);
  if (error) {
    configFile.close();
    return config;
  }

  configFile.close();

  config.ssid = doc["ssid"] | "";
  config.wifiPass = doc["wifiPass"] | "";
  config.latitude = doc["latitude"] | 0.0;
  config.longitude = doc["longitude"] | 0.0;
  config.haKey = doc["haKey"] | "";
  config.homeEntity = doc["homeEntity"] | "";
  config.haUrl = doc["haUrl"] | "";
  config.calUser = doc["calUser"] | "";
  config.calKey = doc["calKey"] | "";
  config.ncUrl = doc["ncUrl"] | "";

  JsonArray calUrls = doc["calUrls"];
  for (JsonVariant url : calUrls) {
    config.calUrls.push_back(url.as<String>());
  }

  return config;
}